from pymarc import (Field, Subfield, Record)
from typing import List, Optional

from rara_tools.constants import EMPTY_INDICATORS
from rara_tools.normalizers.viaf import VIAFRecord
from rara_tools.normalizers import RecordNormalizer

from typing import List


class BibRecordNormalizer(RecordNormalizer):
    """ Normalize bib records. """

    def __init__(self, linking_results: List[dict] = [], sierra_data: List[dict] = [],
                 ALLOW_EDIT_FIELDS: List[str] = ["008", "925"],
                 REPEATABLE_FIELDS: List[str] = ["667"]):
        super().__init__(linking_results, sierra_data)
        self.DEFAULT_LEADER = "00399nz  a2200145n  4500" # must be 24 digits
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS
        
        self.records_extra_data = []
        self.sierra_data = sierra_data
        self.records = self._setup_records(linking_results, sierra_data)

    def _normalize_sierra(self, record: Record) -> Record:
        
        suffix_008 = "|||aznnnaabn          || |||      "
        
        fields = [
            Field(
                tag="008",
                data=f"{self.current_timestamp()}{suffix_008}"
            ),
        ]

        self._add_fields_to_record(record, fields)
        
    def _include_name_variations(self, record: Record, viaf_record: VIAFRecord) -> None:
        """ Include name variations from VIAF record as 400|t fields """
        
        if not viaf_record or not viaf_record.name_variations:
            return
        
        existing_name_variations = record.get_fields("400")
        existing_variations = [sf.value for field in existing_name_variations for sf in field.get_subfields("t")]
        
        fields = []
        
        for variation in viaf_record.name_variations:
            if variation not in existing_variations:
                fields.append(
                    Field(
                        tag="400",
                        indicators=EMPTY_INDICATORS,
                        subfields=[
                            Subfield("t", variation)
                        ]
                    )
                )
        
        self._add_fields_to_record(record, fields)
        
    def _add_author(self, record: Record, viaf_record: Optional[VIAFRecord], original_entity: str) -> Optional[Field]:
        if record.get("100") or record.get("110") or record.get("111"):
            return record

        type_map = {
            "Personal": "100",
            "Corporate": "110",
            "Collective": "111"
        }
        
        tag = type_map.get(getattr(viaf_record, "name_type", None), "100")
        title = getattr(viaf_record, "name", None) or original_entity

        fields = [Field(tag=tag, indicators=EMPTY_INDICATORS, subfields=[Subfield("t", title)])]
        
        self._add_fields_to_record(record, fields)
        
        if viaf_record:
            self._include_name_variations(record, viaf_record)

    def _normalize_viaf(self, record: Record, viaf_record: VIAFRecord, original_entity: str) -> None:
        
        if not viaf_record:
            # viaf record not found, include original entity as 100|t
            self._add_author(record, viaf_record=None, original_entity=original_entity)
            return record

        viaf_id = viaf_record.viaf_id
        fields = [
            Field(
                tag="035",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("a", viaf_id)
                ]
            )
        ]

        self._add_fields_to_record(record, fields)
        self._add_author(record, viaf_record, original_entity=original_entity)

    def _normalize_record(self, record: Record, sierraID: str,
                          viaf_record: VIAFRecord, is_editing_existing_record: bool, original_entity: str) -> Record:

        self._normalize_sierra(record)
        self._normalize_viaf(record, viaf_record, original_entity=original_entity)

        return record



from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient


def test_fetch_clusters_by_id_list():
    viaf_ids = ["7432247", "456"]
    client = VIAFClient()

    results = client.fetch_viaf_clusters(viaf_ids)
    assert len(results) == 2
    assert results["456"] == {}
    assert len(results["7432247"]) > 0


def test_fetch_viaf_results_for_normalizer():
    viaf_ids = ["7432247", "456"]
    client = VIAFClient()

    results = client.get_normalized_data_by_ids(viaf_ids)
    assert len(results) == 2

def test_get_normalized_data_by_search_term():
    client = VIAFClient()
    entity = "Kaja Kallas"

    # Test without verification
    record = client.get_normalized_data_by_search_term(
        search_term=entity,
        verify=False
    )
    assert record.name == "Kallas, Siim"

    record = client.get_normalized_data_by_search_term(
        search_term=entity,
        verify=True
    )
    assert record.name == "Kallas, Kaja"

    record = client.get_normalized_data_by_search_term(
        search_term = "ahsjhajhsjh",
        verify=True,
        max_records=5
    )
    assert record == None

def test_viaf_search_term_query_working():
    client = VIAFClient()
    assert client.check_search_term_query()

def test_viaf_id_query_working():
    client = VIAFClient()
    assert client.check_id_query()


def test_subfield_based_main_field_extraction():
    """ VIAFRecod.name should be retrieved from
    subfield (a) for persons, corporations etc, but
    from field (t) for titles.
    """
    client = VIAFClient()
    record = client.get_normalized_data_by_search_term(
        search_term="Kevade",
        field="local.uniformTitleWorks",
        verify=True
    )
    assert record.name == "Kevade"

    record = client.get_normalized_data_by_search_term(
        search_term="Oskar Luts",
        verify=True
    )
    assert record.name == "Luts, Oskar"

def test_changing_allowed_sources():
    client = VIAFClient(allowed_viaf_sources=["PLWABN"])
    record = client.get_normalized_data_by_search_term(
        search_term="Anora",
        field="local.uniformTitleWorks",
        verify=False
    )
    assert record.name == "Anora (film)"

    client = VIAFClient(allowed_viaf_sources=["LC"])
    record = client.get_normalized_data_by_search_term(
        search_term="Anora",
        field="local.uniformTitleWorks",
        verify=False
    )
    assert record.name == "Anora (Motion picture)"



import pymarc
from typing import NoReturn


class MarcConverter:
    def __init__(self):
        pass

    @staticmethod
    def marc21xml_to_mrc(input_file: str, output_file: str) -> NoReturn:
        """ Converts Marc21XML file into a MRC file.
        """
        with open(output_file, "wb") as f:
            writer = pymarc.MARCWriter(f)
            records = pymarc.marcxml.map_xml(writer.write, input_file)



import os

import pytest
from rara_tools.exceptions import (S3ConnectionException, S3InitException,
                                   S3InputException)
from rara_tools.s3 import S3Files

TEST_URL = os.getenv("S3_URL", "s3.eu-central-003.backblazeb2.com")
TEST_ACCESS_KEY = os.getenv("S3_ACCESS_KEY", "")
TEST_SECRET_KEY = os.getenv("S3_SECRET_KEY", "")
TEST_BUCKET = os.getenv("S3_TEST_BUCKET", "rara-test")


def test_s3_missing_init_params():
    """Tests class init with missing params.
    """
    # Missing URL
    with pytest.raises(S3InitException) as exception_info:
        S3Files()
    assert "URL" in str(exception_info.value)
    # Missing access key
    with pytest.raises(S3InitException) as exception_info:
        S3Files(url="asd")
    assert "access key" in str(exception_info.value)
    # MIssing secret key
    with pytest.raises(S3InitException) as exception_info:
        S3Files(url="asd", access_key="asd")
    assert "secret key" in str(exception_info.value)
    # Missing bucket
    with pytest.raises(S3InitException) as exception_info:
        S3Files(url="asd", access_key="asd", secret_key="asd")
    assert "Bucket" in str(exception_info.value)

def test_s3_incorrect_init_params():
    """Tests class init with incorrect params.
    """
    # Incorrect URL
    with pytest.raises(S3ConnectionException) as exception_info:
        S3Files(
            url="kuhuiganes.com",
            access_key="asd",
            secret_key="asd",
            bucket="asd"
        )
    assert "Error connecting to bucket" in str(exception_info.value)
    assert "Max retries exceeded" in str(exception_info.value)
    # Incorrect access key
    with pytest.raises(S3ConnectionException) as exception_info:
        S3Files(
            url=TEST_URL,
            access_key="asd",
            secret_key="asd",
            bucket="asd"
        )
    assert "Error connecting to bucket" in str(exception_info.value)
    assert "InvalidAccessKeyId" in str(exception_info.value)
    # Incorrect secret key
    with pytest.raises(S3ConnectionException) as exception_info:
        S3Files(
            url=TEST_URL,
            access_key=TEST_ACCESS_KEY,
            secret_key="asd",
            bucket="asd"
        )
    assert "Error connecting to bucket" in str(exception_info.value)
    assert "SignatureDoesNotMatch" in str(exception_info.value)
    # Incorrect bucket
    with pytest.raises(S3ConnectionException) as exception_info:
        S3Files(
            url=TEST_URL,
            access_key=TEST_ACCESS_KEY,
            secret_key=TEST_SECRET_KEY,
            bucket="asd"
        )
    assert "Error connecting to bucket" in str(exception_info.value)
    assert "AccessDenied" in str(exception_info.value)


def test_s3_incorrect_path():
    """Tests class methods with incorrect input paths.
    """
    s3 = S3Files(
            url=TEST_URL,
            access_key=TEST_ACCESS_KEY,
            secret_key=TEST_SECRET_KEY,
            bucket=TEST_BUCKET
    )
    # incorrect upload path
    with pytest.raises(S3InputException) as exception_info:
        s3.upload("alksje")
    assert "does not exist" in str(exception_info.value)



from typing import NoReturn
from collections.abc import Iterator
from rara_tools.parsers.marc_parsers.base_parser import BaseMARCParser
from rara_tools.parsers.marc_records.title_record import TitleRecord
from rara_tools.constants.parsers import LOGGER


class TitlesMARCParser(BaseMARCParser):
    """ MARC parser for titles' .mrc files.
    """
    def __init__(self,
            marc_file_path: str,
            add_variations: bool = True
        ) -> NoReturn:
        """ Initializes OrganizationsMARCParser object.

        Parameters
        -----------
        marc_file_path: str
            Full path to .mrc file containing titles' data.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        super().__init__(
            marc_file_path=marc_file_path,
            add_variations=add_variations
        )

    def record_generator(self) -> Iterator[TitleRecord]:
        """ Generates TitleRecord objects.
        """
        LOGGER.info(
            f"Generating title records from " \
            f"MARC dump '{self.marc_file_path}'."
        )
        for record in self.marc_record_generator():
            title_record = TitleRecord(
                record=record,
                add_variations=self.add_variations
            )
            yield title_record.full_record



import logging
import os
import pathlib
import time
import uuid
from typing import Any, Generator, List, Optional

from minio import Minio, S3Error

from .exceptions import (
    S3ConnectionException,
    S3InitException,
    S3InputException,
    S3DownloadException
)

logger = logging.getLogger("tools.s3")


class S3Files:
    """ A class to manage file uploads and downloads from/to S3.
    """

    def __init__(
            self,
            url: Optional[str] = None,
            access_key: Optional[str] = None,
            secret_key: Optional[str] = None,
            bucket: Optional[str] = None,
            **minio_kwargs: dict[str, Any],
    ):
        if not url:
            raise S3InitException("S3 URL not set!")
        if not access_key:
            raise S3InitException("S3 access key not set!")
        if not secret_key:
            raise S3InitException("S3 secret key not set!")
        if not bucket:
            raise S3InitException("Bucket not set!")
        self.bucket = bucket
        self.minio_client = Minio(
            url,
            access_key=access_key,
            secret_key=secret_key,
            **minio_kwargs
        )
        # Check S3 connection
        try:
            self.minio_client.bucket_exists(bucket)
        except Exception as e:
            raise S3ConnectionException(f"Error connecting to bucket: {e}")

    def _put_file(self, file_path: str, s3_path_name: str):
        if not os.path.exists(file_path):
            raise S3InputException(f"File '{file_path}' does not exist in file system!")
        return self.minio_client.fput_object(self.bucket, s3_path_name, file_path)

    def list(self, prefix: Optional[str] = "", recursive: Optional[bool] = True) -> List:
        """Lists all available directories or files in S3 bucket.
        :param: prefix str: Limits the listing to a given prefix.
        :param: recursive bool: List files recursively.
        :return: List of file paths in S3.
        """
        list_of_objects = self.minio_client.list_objects(self.bucket, prefix=prefix, recursive=recursive)
        list_of_objects = [o.object_name for o in list_of_objects]
        return list_of_objects

    def delete(self, path: str) -> bool:
        """Deletes file in S3.
        :param: path str: Path of the file in S3 to be deleted.
        :return: True.
        """
        list_of_objects = self.minio_client.list_objects(self.bucket, prefix=path, recursive=True)
        list_of_objects = [o.object_name for o in list_of_objects]
        for path_to_delete in list_of_objects:
            self.minio_client.remove_object(self.bucket, path_to_delete)
        return True

    def download(self, path: str, download_dir: Optional[str] = ".") -> Generator[str, str, str]:
        """Downloads file or folder from S3.
        :param: path str: Path to the file or folder in S3.
        :param: download_dir str: Directory to download the files and folders into.
        :return: Generate listing of local paths of the downloaded files. 
        """
        list_of_objects = list(self.minio_client.list_objects(self.bucket, prefix=path, recursive=True))
        for minio_object in list_of_objects:
            full_path = os.path.join(download_dir, minio_object.object_name)
            self._download_file(minio_object.object_name, full_path)
            yield full_path

    def _download_file(self, path, download_dir=".", max_retries=3) -> str:
        """Download a single file with retry and resume support."""
        attempts = 0

        while attempts < max_retries:
            try:
                stat = self.minio_client.stat_object(self.bucket, path)
                file_size = stat.size
                temp_path = download_dir + ".part"
                pathlib.Path(temp_path).parent.mkdir(parents=True, exist_ok=True)

                # Check if a partial file exists
                downloaded_size = os.path.getsize(temp_path) if os.path.exists(temp_path) else 0

                if downloaded_size >= file_size:
                    os.rename(temp_path, download_dir)  # Rename to final filename
                    logger.info(f"Completed: {path}")
                    return str(pathlib.Path(download_dir) / path)

                logger.info(f"Downloading {path} ({downloaded_size}/{file_size} bytes)...")

                # Open file in append mode to resume download
                with open(temp_path, "ab") as f:
                    response = self.minio_client.get_object(self.bucket, path, offset=downloaded_size)
                    for data in response.stream(32 * 1024):  # 32KB chunks
                        f.write(data)
                    response.close()
                    response.release_conn()

                os.rename(temp_path, download_dir)  # Rename temp to final
                logger.info(f"Downloaded: {path}")
                return str(pathlib.Path(download_dir) / path)

            except S3Error as e:
                logger.info(f"Error downloading {path}, attempt {attempts + 1}: {e}")
                attempts += 1
                time.sleep(2 ** attempts)  # Exponential backoff

        raise S3DownloadException(f"Failed to download {path} after {max_retries} attempts.")

    def upload(self, path: str, prefix: Optional[str] = "") -> str:
        """Uploads file or folder to S3 bucket.
        :param: path str: Path to the file to upload in local file system.
        :param: prefix str: Optional prefix for S3 path.
        :returns: File path of the uploaded file or folder in S3.
        """
        # Manage directories
        if os.path.isdir(path):
            s3_path_prefix = f"{prefix}{uuid.uuid4().hex}"
            for root, _, files in os.walk(path):
                for file in files:
                    relative_prefix = root.removeprefix(path).lstrip("/")
                    file_path = os.path.join(root, file)
                    extension = file.split(".")[-1]
                    s3_path_name = os.path.join(s3_path_prefix, relative_prefix, file)
                    self._put_file(file_path, s3_path_name)
            return s3_path_prefix
        # Manage single files
        else:
            path_name = os.path.split(path)[-1]
            extension = path_name.split(".")[-1]
            s3_path_name = f"{prefix}{uuid.uuid4().hex}.{extension}"
            self._put_file(path, s3_path_name)
            return s3_path_name



import os
import json


from rara_tools.constants import YYMMDD_FORMAT, YY_DD_FORMAT
from rara_tools.normalizers import (BibRecordNormalizer, AuthoritiesRecordNormalizer)
from tests.test_utils import (get_linker_res_example, get_formatted_sierra_response, search_viaf_record,
                        check_record_tags_sorted, check_no_dupe_tag_values, check_record_tags_have_values)

from rara_tools.constants.linker import EntityType

from datetime import datetime

from pymarc import Record, JSONReader


TEST_LEVEL = os.getenv("TEST_LEVEL", "unit")

EMPTY_SIERRA_RECORDS = [
    {
        "sierraID": "1",
        "leader": "00000nz  a2200000n  4500",
        "fields": []
    },
]

REQUIRED_FIELDS = ["667", "925"]  # always included after normalization
MOCK_LINKER_ONE_FOUND = get_linker_res_example(
    "oneFound.json")
MOCK_LINKER_MULTIPLE_FOUND = get_linker_res_example(
    "multipleFound.json")
MOCK_LINKER_NOT_FOUND = get_linker_res_example(
    "notFound.json")


def test_normalizers_OK():
    """ Test field editing logic & internals """

    linking_results = [MOCK_LINKER_ONE_FOUND,
                       MOCK_LINKER_MULTIPLE_FOUND]

    test_sierra_data = get_formatted_sierra_response("authorities.json")

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results,
        sierra_data=test_sierra_data,
    )
    
    assert len(normalizer.records_extra_data) == len(normalizer.data)

    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
        sierra_data=test_sierra_data,
    )
    assert len(normalizer.records_extra_data) == len(normalizer.data)

    data = [
        {
            "sierraID": "1",
            "leader": "00000nz  a2200000n  4500",
            "fields": [
                {
                    "667": {
                        "ind1": " ",
                        "ind2": " ",
                        "subfields": [
                            {
                                "a": "Val"
                            }
                        ]
                    }
                },
            ]
        },
    ]

    # default behavior - added if not in record &
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=[],
    )
    for r in normalizer:
        assert r.get_fields("667")[0].get_subfields("a")[0] == "Val"

    # not edited if exists
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=[]
    )
    for r in normalizer:
        assert r.get_fields("667")[0].get_subfields("a")[0] == "Val"

    # allow repeatable, new field will be added
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=["667"]
    )
    for r in normalizer:
        fields_667 = r.get_fields("667")
        assert len(fields_667) == 2
        assert fields_667[0].get_subfields("a")[0] == "Val"
        assert fields_667[1].get_subfields("a")[0] == "Muudetud AI poolt"

    # allow editing, field will be edited
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=["667"],
        REPEATABLE_FIELDS=[]
    )
    for r in normalizer:
        fields_667 = r.get_fields("667")
        assert len(fields_667) == 1
        assert fields_667[0].get_subfields("a")[0] == "Muudetud AI poolt"
        
def _validate_existing_record_notes(record: Record):
    """ Validate notes in existing record - 667 & 925 fields """
    
    fields_667 = record.get_fields("667")
    assert any("Muudetud AI poolt" in f.get_subfields("a") for f in fields_667)
    
    fields_925 = record.get_fields("925")
    assert len(fields_925) == 1
    assert len(fields_925[0].get_subfields("p")) == 1
    assert len(fields_925[0].get_subfields("t")) == 0
    
def _validate_new_record_notes(record: Record):
    """ Validate notes in new record - 667 & 925 fields """
    
    fields_667 = record.get_fields("667")
    
    assert any("Loodud AI poolt" in f.get_subfields("a") for f in fields_667)
    
    fields_925 = record.get_fields("925")
    assert len(fields_925) == 1
    assert len(fields_925[0].get_subfields("p")) == 0
    assert len(fields_925[0].get_subfields("t")) == 1

def validate_bibrecord_normalized(record: Record, has_viaf_data=False):
    """ Validate specific fields in normalized bibrecord """
       
    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40
    # pos 00-05 is current date in YYMMDD format
    timestamp = field_008[0:6]
    # check timestamp in correct format
    try:
        datetime.strptime(timestamp, YYMMDD_FORMAT)
    except ValueError:
        raise AssertionError(f"008 field timestamp {timestamp} is not in format {YYMMDD_FORMAT}")
    
    expected_suffix_008 = "|||aznnnaabn          || |||" + 6 * " "
    suffix = field_008[6:]
    assert suffix == expected_suffix_008
    
    _validate_existing_record_notes(record)
    
def validate_authorities_record_normalized(record: Record, has_viaf_data=False):
    """ Validate specific fields in normalized authority record """
     
    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40
    # pos 00-05 is current date in YYMMDD format
    timestamp = field_008[0:6]
    # check timestamp in correct format
    try:
        datetime.strptime(timestamp, YYMMDD_FORMAT)
    except ValueError:
        raise AssertionError(f"008 field timestamp {timestamp} is not in format {YYMMDD_FORMAT}")
    
    expected_suffix_008 = "|n|adnnnaabn          || |a|" + 6 * " "
    suffix = field_008[6:]
    assert suffix == expected_suffix_008

    _validate_existing_record_notes(record)

    # check that a, b & c subfields have values (can have default or unique)
    field_040_subfields = record.get_fields("040")[0]
    assert len(field_040_subfields.get_subfields("a")) > 0
    assert len(field_040_subfields.get_subfields("b")) > 0
    assert len(field_040_subfields.get_subfields("c")) > 0
    
    # check that 008 field has a value of length 40
    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40

    if has_viaf_data:
        field_043 = record.get_fields("043")[0].get_subfields(
            "c")[0]  # check that 043 has subfield c with value "ee"
        assert field_043 == "ee"

        field_024 = record.get_fields("024")
        for f in field_024:
            assert len(f.get_subfields("0")) > 0  # VIAF url

        field_046 = record.get_fields("046")[0]
        assert len(field_046.get_subfields("f")) > 0  # birth date
        assert len(field_046.get_subfields("g")) > 0  # death date

def test_missing_fields_created_bibrecord_normalization():
    linking_results = [MOCK_LINKER_ONE_FOUND]

    normalizer_entities_only = BibRecordNormalizer(
        linking_results=linking_results,
    )

    normalizer_sierra_data_only = BibRecordNormalizer(
        sierra_data=EMPTY_SIERRA_RECORDS,
    )

    for record in normalizer_entities_only:
        check_record_tags_have_values(
            record, ["008",  # Sierra related, always with bibs
                     "035", "100",  # VIAf enriched
                     ] + REQUIRED_FIELDS
        )
        validate_bibrecord_normalized(record, has_viaf_data=True)

    for record in normalizer_sierra_data_only:
        check_record_tags_have_values(
            record, ["008"  # Sierra related, always with bibs
                     ] + REQUIRED_FIELDS)
        validate_bibrecord_normalized(record)


def test_missing_fields_created_authorities_normalization():

    linking_results = [MOCK_LINKER_ONE_FOUND]

    normalizer_entities_only = AuthoritiesRecordNormalizer(
        linking_results=linking_results,  # find one match
    )

    normalizer_sierra_data_only = AuthoritiesRecordNormalizer(
        sierra_data=EMPTY_SIERRA_RECORDS,
    )

    for r in normalizer_entities_only:
        check_record_tags_have_values(r, ["008", "040",  # SIERRA related
                                          "024", "043", "046"  # VIAF enriched
                                          ] + REQUIRED_FIELDS)

        validate_authorities_record_normalized(r, True)

    for r in normalizer_sierra_data_only:
        check_record_tags_have_values(
            r, ["040"] + REQUIRED_FIELDS)
        validate_authorities_record_normalized(r)


def test_normalized_fields_sorted():

    unsorted_bibdata = [
        {
            "sierraID": "1",
            "leader": "00000nz  a2200000n  4500",
            "fields": [
                {
                        "035": {
                            "ind1": " ",
                            "ind2": " ",
                            "subfields": [
                                {
                                    "a": "(ErESTER)<1>"
                                }
                            ]
                        }
                },
                {
                    "008": "220805|||aznnnaabn          || |||      nz n  "
                },
                {
                    "046": {
                        "ind1": " ",
                        "ind2": " ",
                        "subfields": [
                            {
                                "k": "1912"
                            }

                        ]
                    }
                },
            ]
        }
    ]

    normalizers = (BibRecordNormalizer, AuthoritiesRecordNormalizer)

    for normalizer in normalizers:
        normalizer = normalizer(
            linking_results=[],
            sierra_data=unsorted_bibdata
        )

        for r in normalizer:
            check_no_dupe_tag_values(r)
            check_record_tags_sorted(r)


def test_authority_normrecord_found_in_es_and_normalized():
    """ KATA elastic normkirjete seast leitakse 1 vaste & normaliseerija täiendab leitud normkirjet VIAF infoga.
        - valideeri normaliseerimise mapping, mis autori tabelis. Täiendatud väljad ja VIAFist info
        - Valideeri märge lisatud (TODO) """
    # Presume, author name identified and sent to linker
    linker_res = get_linker_res_example(
        "oneFound.json")  # single result
    linking_results = [linker_res]

    # 1 result found
    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results
    )

    data = normalizer.data

    assert len(data) == 1
    record = next(iter(JSONReader(
            json.dumps(normalizer.data, ensure_ascii=False)
    )))
    check_record_tags_have_values(record, ["040"] + REQUIRED_FIELDS)
    validate_authorities_record_normalized(record, has_viaf_data=True)
    
def _run_normalizer(linked_data):
    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linked_data
    )
    return normalizer.data

def test_normalizer_handles_bad_inputs():
    linker_res = get_linker_res_example(
        "oneFound.json")
    
    # pop the leader field to simulate record without leader
    linker_res["linked_info"][0]["json"].pop("leader", None)
    _run_normalizer([linker_res])

    # make fields empty to simulate a record with no fields
    linker_res["linked_info"][0]["json"]["fields"] = []
    _run_normalizer([linker_res])

    # pop the fields to simulate a record with no fields
    linker_res["linked_info"][0]["json"].pop("fields", None)
    _run_normalizer([linker_res])

    inputs = ["", None, [], {}, 123]
    
    _run_normalizer(inputs)

def test_matching_sierra_record_viaf_id_found():
    """normkirjelt leitakse VIAF ID, vajadusel normi asukoht, kus see ID sisaldub."""
    pass


def test_matching_sierra_record_viaf_id_not_found():
    """kirjelt VIAF IDd ei leitud, soorita otsing VIAFi pihta, et leida _vastutav isik_?. Loo uus vastavalt otsingu tulemusele."""
    pass


def test_authorities_normalizer_checks():
    """
    - kontrolli kas tuvastatud nimi on SIERRAst leitud vaste 1XX, 4XX väljadel. Kui pole, siis lisa 4XX väljale.
    - kontrolli, kas VIAF andmete nimekujud on normkandes olemas. Kui pole, lisa need 4XX väljale.
    - Kontrolli, kas VIAF kandes on sünni ja surma daatumid ja kas need klapivad normkandes olevaga. Kui pole, siis liiguta normkandest kogu 1XX väli 4XX väljale. Seejärel loo uute daatumitega 1XX väli.
    - Kontrolli, et väljal 046 olevad daatumid klapiksid just 1xx väljale lisatuga. Kui andmeid muudeti, siis märgi, et baasis on normkanne muutunud
    """
    pass

def test_add_birth_and_death_dates():
    """ Test adding birth & death dates to 046 field from VIAF record. Expected Date format YYMMDD """
    
    # Case one: viaf record has both birth & death date
    normalizer = AuthoritiesRecordNormalizer()
    record = Record()

    viaf_record = normalizer._get_viaf_record(
        record,
        entity="Paul Keres"
    )
    for value in [viaf_record, viaf_record.birth_date, viaf_record.death_date]:
        assert value is not None
    
    normalizer._add_birth_and_death_dates(record, viaf_record)
    
    field_046 = str(record.get_fields("046")[0])
    assert field_046 == "=046  \\\\$f19160107$g19750605"
    
    # Case two: viaf record has birth date, but no death date (author still alive)
    viaf_record = normalizer._get_viaf_record(
        record,
        entity="Andrus Kivirähk"
    )
    
    assert viaf_record is not None
    assert viaf_record.birth_date is not None
    assert viaf_record.death_date is 0 
    
    record = Record()    
    normalizer._add_birth_and_death_dates(record, viaf_record)
    
    field_046 = str(record.get_fields("046")[0])
    # empty indicators represented with \
    assert field_046 == "=046  \\\\$f19700817"
    
    # Case 3 - viaf record has no birth or death date
    viaf_record = normalizer._get_viaf_record(
        record,
        entity="Eesti Interlingvistika Selts"      
    )
    record = Record()
    normalizer._add_birth_and_death_dates(record, viaf_record)
    # should not add 046 field
    fields_046 = record.get_fields("046")
    assert len(fields_046) == 0
    
def test_add_nationality():
    """ Test adding nationality from VIAF record to 043 field """
    
    # Case 1: nationality is not estonian - do not add 043 field
    
    normalizer = AuthoritiesRecordNormalizer()
    record = Record()
    
    viaf_record = normalizer._get_viaf_record(
        record,
        entity="Paulo Coelho"
    )
    
    assert viaf_record is not None
    assert viaf_record.nationality == "br"

    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 0 
    
    # Case 2, nationality missing
    
    normalizer = AuthoritiesRecordNormalizer()
    record = Record()
    
    viaf_record = normalizer._get_viaf_record(
        record,
        entity="123zsxcasd"
    )
    
    assert viaf_record is None
    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 0

    # Case 3 - Nationality is Estonian - add 043 field
    viaf_record = normalizer._get_viaf_record(
        record,
        entity="Jaan Tõnisson"
    )

    assert viaf_record is not None
    assert viaf_record.nationality == "ee"

    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 1
    assert str(fields_043[0]) == "=043  \\\\$cee"
    
    # Case 4 - 043 field already exists - should not get edited (not in ALLOW_EDIT_FIELDS)
    linking_results = [{
        "original_entity": "Eduard Vilde",
        "entity_type": EntityType.PER,
        "linked_info": [
            {
                "json": {
                    "leader": "00736nz  a2200217n  4500",
                    "fields": [
                        {
                            "043": {
                                "ind1": " ",
                                "ind2": " ",
                                "subfields": [
                                    {
                                        "c": "gb"
                                    }
                                ]
                            }
                        }
                    ]
                }
    
            }
        ]
    }]
    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results,
    )
    record = normalizer.first
    # mock run add nationality with foreign VIAF record
    viaf_record = normalizer._get_viaf_record(
        record,
        entity="Jaan Tõnisson"
    )
    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 1
    assert str(fields_043[0]) == "=043  \\\\$cgb"
    

def test_create_new_normrecord():
    
    # Case 1 - one entity found - update existing norm record    
    linker_res = get_linker_res_example(
        "oneFound.json")
    linking_results = [linker_res]

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results
    )
    data = normalizer.data

    assert len(data) == 1  # should update existing record

    # Case 2 - entities not found - create new norm record
    linking_results = [MOCK_LINKER_NOT_FOUND]
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 1
    
    record = normalizer.first
    leader = str(record.leader)
    assert leader == "01682nz  a2200349n  4500"
    assert len(leader) == 24
    _validate_new_record_notes(record)

    # Case 3 - multiple entities found - create new norm record
    linker_res = get_linker_res_example(
        "multipleFound.json")
    linking_results = [linker_res]
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 1
    record = normalizer.first
    
    _validate_new_record_notes(record)
    # validate leader
    leader = str(record.leader)
    assert leader == "01682nz  a2200349n  4500"
    assert len(leader) == 24
    
    # Case 3 with bibnormalizer - update with better example
    linker_res = get_linker_res_example(
        "multipleFound.json")
    linking_results = [linker_res]

    normalizer = BibRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 1
    record = normalizer.first
    
    _validate_new_record_notes(record)
    # validate leader
    leader = str(record.leader)
    assert leader == "00399nz  a2200145n  4500"
    assert len(leader) == 24
    
    # Mock not found title
    linking_results = [{
        "original_entity": "Eesti Ekspress",
        "entity_type": EntityType.TITLE,
        "linked_info": []
    }]
    normalizer = BibRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 1
    record = normalizer.first
    
def test_680_field_on_existing_record_moved_to_667():
    """ 680 Should not be added for new, if exists on existing record, should be moved to 667 """
    linker_res = get_linker_res_example(
    "oneFound.json")
    
    # add 680 field to existing record will have two 680 fields after adding
    linker_res["linked_info"][0]["json"]["fields"].append({
        "680": {
            "ind1": " ",
            "ind2": " ",
            "subfields": [
                {"a": "Test 680 field" }
            ]
        }
    })
    initial_record = JSONReader(
            json.dumps([linker_res["linked_info"][0]["json"]], ensure_ascii=False)
        )
    record = next(iter(initial_record))
    fields_680 = record.get_fields("680")
    assert len(fields_680) == 2  # inital plus new
    fields_667 = record.get_fields("667")
    assert len(fields_667) == 0  # initial only

    linking_results = [linker_res]

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results
    )
    record = normalizer.first
    
    fields_680 = record.get_fields("680")
    assert len(fields_680) == 0
    fields_667 = record.get_fields("667")
    assert len(fields_667) == 3  # original + moved from 680 + new note    

def test_date_formatting():
    normalizer = AuthoritiesRecordNormalizer()
    
    dates = {
        "19700712": "19700712",
        "1970": "1970",
        "1970-07": "197007",
        "2001-12-31": "20011231",
        "1999-01": "199901",
    }
    
    for input_date, expected in dates.items():
        assert normalizer._format_date(input_date) == expected
    
    # invalid date formats - should return empty string
    invalid_dates = ["abcd", "199A0101"]
    for date in invalid_dates:
        assert normalizer._format_date(date) == ""
        
def test_new_bibrecord_title_included():
    """ normrecord for bibs has to always have the 1XX|t field filled """
    
    # Case 1 No linker response, & Viaf record found
    linking_results = [{
        "original_entity": "Lord of the Rings",
        "entity_type": EntityType.TITLE,
        "linked_info": []
    }]

    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
    )
    data = normalizer.data
    assert len(data) == 1  # should enrich existing record
    record = normalizer.first
    
    _validate_new_record_notes(record)
    fields_100 = record.get_fields("100")
    assert len(fields_100) == 1
    assert fields_100[0].get_subfields("t")[0] == "Lord of the rings"
    
    # Case 2 - Viaf record not found - should use original entity
    linking_results = [{
        "original_entity": "Roolijoodiku katastroofiline jõulusõit",
        "entity_type": EntityType.TITLE,
        "linked_info": []
    }]
    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
    )
    record = normalizer.first
    data = normalizer.data
    assert len(data) == 1  # should enrich existing record
    
    fields_100 = record.get_fields("100")
    assert len(fields_100) == 1
    assert fields_100[0].get_subfields("t")[0] == "Roolijoodiku katastroofiline jõulusõit"



from typing import List, NoReturn
from pymarc.record import Record
from pymarc import MARCReader
from abc import abstractmethod
from collections.abc import Iterator, Iterable
import jsonlines


class BaseMARCParser:
    """ Base class for MARC parsers.
    """
    def __init__(self,
            marc_file_path: str,
            add_variations: bool = True
        ) -> NoReturn:
        """ Initializes BaseMARCParser object.

        Parameters
        -----------
        marc_file_path: str
            Full path to .mrc file containing EMS data.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        self.add_variations = add_variations
        self.marc_file_path = marc_file_path

    def _write_line(self, line: dict, file_path: str) -> NoReturn:
        with jsonlines.open(file_path, "a") as f:
            f.write(line)

    def marc_record_generator(self) -> Iterator[Record]:
        """ Generates pymarc.record.Record objects.
        """
        with open(self.marc_file_path, "rb") as fh:
            reader = MARCReader(fh)
            for record in reader:
                if record:
                    yield record

    @abstractmethod
    def record_generator(self) -> Iterator:
        pass

    def save_as_jl(self, jl_file_path: str) -> NoReturn:
        for record in self.record_generator():
            self._write_line(record, jl_file_path)



from rara_tools.constants.linker import (
    LOGGER, URLSource, KeywordSource, EntityType, KeywordType, KeywordMARC,
    KEYWORD_MARC_MAP,  KEYWORD_TYPES_TO_IGNORE, KEYWORD_TYPE_MAP,
    EMS_ENTITY_TYPES, SIERRA_ENTITY_TYPES, UNLINKED_KEYWORD_MARC_FIELD,
    URL_SOURCE_MAP
)
from rara_tools.core_formatters.formatted_object import FormattedObject
from typing import List, Dict, NoReturn, Tuple, Any

class FormattedKeyword(FormattedObject):
    def __init__(self, object_dict: dict, linked_doc: Any,
            main_taxnomy_lang: str, url_source_map: str = URL_SOURCE_MAP
    ) -> NoReturn:
        super().__init__(
            object_dict=object_dict,
            linked_doc=linked_doc,
            original_entity_key="keyword"
        )

        self.main_taxnomy_lang: str = main_taxnomy_lang

        self.original_keyword: str = self.original_entity
        self.score: float = self.object_dict.get("score")
        self.count: int = self.object_dict.get("count")
        self.method: str = self.object_dict.get("method")
        self.model_arch: str = self.object_dict.get("model_arch", self.method)
        self.keyword_type: str = self.object_dict.get("entity_type")
        self.article_id: str | None = self.object_dict.get("article_id", None)

        self.entity_type: str = KEYWORD_TYPE_MAP.get(self.keyword_type, "")
        self.url_source_map: dict = url_source_map

        self.__keyword_source: str = ""
        self.__indicator_1: str = ""
        self.__indicator_2: str = ""
        self.__url: str | None = None
        self.__url_source: str | None = None
        self.__marc_field: str = ""

        self.__language: str = ""
        self.__author: str | None = None


    @property
    def keyword(self) -> str:
        return self.entity

    @property
    def keyword_source(self) -> str:
        if not self.__keyword_source:
            if not self.is_linked:
                source = KeywordSource.AI
            elif self.entity_type in EMS_ENTITY_TYPES:
                source = KeywordSource.EMS
            elif self.entity_type in SIERRA_ENTITY_TYPES:
                if self.linked_doc and self.linked_doc.elastic:
                    source = KeywordSource.SIERRA
                elif self.linked_doc and self.linked_doc.viaf:
                    source = KeywordSource.VIAF
                else:
                    source = KeywordSource.AI
            else:
                source = KeywordSource.AI
            self.__keyword_source = source
        return self.__keyword_source

    @property
    def indicator1(self) -> str:
        if not self.__indicator_1:
            ind1, ind2 = self._get_indicators()
            self.__indicator_1 = ind1
            self.__indicator_2 = ind2
        return self.__indicator_1

    @property
    def indicator2(self) -> str:
        if not self.__indicator_2:
            ind1, ind2 = self._get_indicators()
            self.__indicator_1 = ind1
            self.__indicator_2 = ind2
        return self.__indicator_2

    @property
    def url(self) -> str:
        if self.__url == None:
            url_info = self._get_url_info()
            self.__url = url_info.get("url")
            self.__url_source = url_info.get("url_source")
        return self.__url

    @property
    def url_source(self) -> str:
        if self.__url_source == None:
            url_info = self._get_url_info()
            self.__url = url_info.get("url")
            self.__url_source = url_info.get("url_source")
        return self.__url_source

    @property
    def marc_field(self) -> int:
        if not self.__marc_field:
            # TODO: teoste + isikute loogika!!!!
            if self.is_linked:
                marc_field = KEYWORD_MARC_MAP.get(str(self.keyword_type), "")
            else:
                marc_field = UNLINKED_KEYWORD_MARC_FIELD

            if self.entity_type == EntityType.TITLE:
                if self.author:
                    marc_field = KeywordMARC.TITLE_LINKED
                else:
                    marc_field = KeywordMARC.TITLE
            self.__marc_field = marc_field
        return self.__marc_field


    @property
    def persons_title(self) -> str:
        return self.titles


    @property
    def language(self) -> str:
        if not self.__language:
            if self.is_linked:
                self.__language = self.main_taxnomy_lang
            else:
                self.__language = self.object_dict.get("language", "")
        return self.__language

    @property
    def author(self) -> str:
        # Only relevant for titles!
        if self.__author == None:
            self.__author = ""
            if self.entity_type == EntityType.TITLE:
                if self.original_record:
                    self.__author = self.original_record.author_name
                elif self.viaf_info:
                    pass
                    #self.__author = self.viaf_info.get
        return self.__author



    def _get_url_info(self) -> dict:
        """ Finds URL identifier from LinkedDoc based on
        given entity type.

        Parameters
        -----------
        linked_doc: LinkedDoc | None
            A LinkedDoc class instance.
        entity_type: str
            Entity type for detecting correct URL source.

        Returns
        ----------
        dict:
            Dictionary with keys `url` - URL identifier and
            `url_source` - source of the URL (e.g. "EMS").

        """
        url_source = self.url_source_map.get(self.entity_type, "")
        url = ""

        if self.linked_doc:
            if url_source == URLSource.EMS:
                url = self.linked_doc.elastic.get("ems_url", "")
            elif url_source == URLSource.VIAF:
                url = self.viaf_info.get("viaf_url", "")
        if not url:
            url_source = ""

        url_info = {"url": url, "url_source": url_source}

        LOGGER.debug(
            f"Detected URL info: {url_info}. Used entity_type = {self.entity_type}. " \
            f"URL source map = {self.url_source_map}."
        )
        return url_info

    def _get_indicators(self) -> Tuple[str, str]:
        """ Find MARC indicators 1 and 2.
        """
        ind1 = " "
        ind2 = " "
        if self.entity_type in SIERRA_ENTITY_TYPES:
            if self.entity_type == EntityType.PER:
                if "," in self.keyword:
                    ind1 = "1"
                else:
                    ind1 = "0"
            elif self.entity_type == EntityType.ORG:
                # 1 märksõna esimeseks elemendiks võimupiirkonna nimi, nt:
                #    (a) Eesti (b) Riigikogu - raske automaatselt määrata
                # 2 märksõna esimeseks elemendiks nimi pärijärjestuses
                ind1 = "2"
            else:
                ind1 = "0"

            if not self.is_linked:
                ind2 = "4"
        elif self.entity_type in EMS_ENTITY_TYPES:
            ind2 = "4"
        return (ind1, ind2)


    def to_dict(self) -> dict:
        keyword_dict = {
            "count": self.count,
            "dates": self.dates,
            "entity_type": self.keyword_type,
            "indicator1": self.indicator1,
            "indicator2": self.indicator2,
            "is_linked": self.is_linked,
            "keyword": self.keyword,
            "keyword_source": self.keyword_source,
            "lang": self.language,
            "location": self.location,
            "marc_field": self.marc_field,
            "method": self.method,
            "model_arch": self.model_arch,
            "numeration": self.numeration,
            "organisation_sub_unit": self.organisation_sub_unit,
            "original_keyword": self.original_keyword,
            "persons_title": self.persons_title,
            "score": self.score,
            "url": self.url,
            "url_source": self.url_source,
            "author": self.author,
            "article_id": self.article_id
        }
        return keyword_dict
